/*********************************************************************
 Copyright (C) 2003  Imperial College London
 Multimedia Knowledge Management Group

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*********************************************************************/

#ifndef ANNOTATE_H
#define ANNOTATE_H

#include <jpeglib.h>

#define DEBUG pars.flag.d

typedef struct {
    int w, h, q;                /* Width, Heigth */
    enum type { colour, bw } t; /* type is unused, all images are RGB */
    JSAMPLE *cont;              /* w*h*3 consecutive intensities */
    
    int blocksW, blocksH;
    JSAMPLE  *dctCOEF;
} AImage;


char *ftfname(char *jpgname, int ti, int tj);
AImage ReadImageFile(char *name);
AImage sample(AImage im, int k);
AImage sampleDCT(AImage im, int nBlocks);
AImage ConvertToGrey(AImage im, int quant);
void WriteImageFile(AImage im, FILE * fp);
FILE *ftfopen(char *name);
void ftfclose(FILE * fp);

#define efree(p) { if((p)) free((p)), (p)=0; }
void *clrmalloc(unsigned int size);
void *erealloc(void *p, unsigned int size);
char *mystrdup(char *p);

void fatal();
int warning();

char *ints2str(int a, int b, int c, int d, int e);

void recursive(int argc, char **argv, char *e);

int externalprog(FILE * fp, AImage im);

int echo(FILE * fp, AImage im);
int mksqthumb(FILE * fp, AImage im), mksqthumbcheck();
int thumbnail(FILE * fp, AImage im), thumbnailcheck(void);
int convolution(FILE * fp, AImage im), convolutioncheck(void);

int smoothness(FILE * fp, AImage im);
int uniformity(FILE * fp, AImage im);
int variance(FILE * fp, AImage im), variancecheck(void);

int rgb(FILE * fp, AImage im), XYZcheck(void);
int hsvlin(FILE * fp, AImage im);
int hsvvol(FILE * fp, AImage im);
int hsllin(FILE * fp, AImage im);
int hslvol(FILE * fp, AImage im);
int ypcbcr(FILE * fp, AImage im);
int cieluv(FILE * fp, AImage im);
int cielab(FILE * fp, AImage im);
int hds(FILE * fp, AImage im), hdscheck(void);
int margrgb(FILE * fp, AImage im), margXYZcheck(void);
int marghsvlin(FILE * fp, AImage im);
int marghsvvol(FILE * fp, AImage im);
int marghsllin(FILE * fp, AImage im);
int marghslvol(FILE * fp, AImage im);
int margypcbcr(FILE * fp, AImage im);
int margcieluv(FILE * fp, AImage im);
int margcielab(FILE * fp, AImage im);

int tamuraS(FILE * fp, AImage im), tamuraCheck(void);
int tamuraImage(FILE * fp, AImage im), tamuraImageCheck(void);
int histTamuraCheck(void);      /* tamura(FILE *fp, AImage im), same as rgb */
int margTamuraCheck(void);      /* margTamura(FILE *fp, AImage im), same as margrgb */

int co_occurrence(FILE * fp, AImage im), co_occ_check(void);
int co_occurrence_sum(FILE * fp, AImage im);

int gabor(FILE * fp, AImage im), gaborCheck(void);
int histGabor(FILE * fp, AImage im), winGaborCheck(void);

int dctCoefs(FILE * fp, AImage im), dctCoefsCheck(void);

int wavelet(FILE * fp, AImage im), waveletCheck(void);

AImage RGBtoYpCbCr(AImage im);
AImage convertToTamura(AImage im, int c, int n, int d);

char *xgetline(FILE *fp, char *buf, size_t *linenum);

typedef struct {
    char *fname, *scanline, *printline, *dfault;
    int (*fct) (FILE * fp, AImage im), (*fctcheck) ();
} Annotations;


#define MAXPARAMS 7
typedef struct {
    char space[MAXPARAMS][16];
    int number[MAXPARAMS];
    char what[MAXPARAMS];
    void *s[MAXPARAMS];
    int newHeigth;
    int newWidth;
    
    struct {
        int r, z, d, n, U, C, G, f, s, q, p, t[3], dct;
    } flag;
    int destsize, maskbottom, useYCbCr, tx, ty, gridG, gridS, nblocks, sigma, memleft, useTamura, win, k, convtype;

    Annotations *ap;
    int (*todo) (FILE * fp, AImage im);
    char *jpgname, *ftfname, *ftfext, *imageext;
} Params;

extern Params pars;


typedef struct {
    int N, r, g, b, index;
    double sum;
    struct {
        char dr, dg, db;        // int dindex;
        float w;
    } c[0];
} Coord;

Coord *smoothing_G(int r, int g, int b), *smoothing_S(int r, int g, int b);

#endif
