/*********************************************************************
 Copyright (C) 2003  Imperial College London
 Multimedia Knowledge Management Group

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*********************************************************************/

/*
 * dctcoefs.c
 * 
 * Joao Magalhaes, 12 April 2005 
 * Computes the DCT coefficients and outputs the selected number of 
 * components.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <errno.h>
#include <ctype.h>
#include <string.h>

#include "annotate.h"

#define DCTSIZE 8

void dct_float(float *data);
void copy_coefs(float *, float *, int);

int dctCoefsCheck()
{
    if (pars.number[0] <= 0 || pars.number[0] > 64)
        fatal("The number of DCT coefficients must be between 1 and 64");
    if (pars.number[1] <= 7 || pars.number[1] > 256)
        fatal("The number of DCT blocks must be between 8 and 256");
        
    pars.flag.dct = 1;
    
    return 1;
}


int dctCoefs(FILE * fp, AImage im)
{
//    float data[DCTSIZE][DCTSIZE] = {
//        {139, 144, 149, 153, 155, 155, 155, 155},
//        {144, 151, 153, 156, 159, 156, 156, 156},
//        {150, 155, 160, 163, 158, 156, 156, 156},
//        {159, 161, 162, 160, 160, 159, 159, 159},
//        {159, 160, 161, 162, 162, 155, 155, 155},
//        {161, 161, 161, 161, 160, 157, 157, 157},
//        {162, 162, 161, 163, 162, 157, 157, 157},
//        {162, 162, 161, 161, 163, 158, 158, 158}
//    };
    float block[DCTSIZE][DCTSIZE];
    int i, j, iBlk, jBlk, blkPosArray, ijPosArray = 0, iPos, jPos;
    int nCoefs, nBlocks;
    float *dctAnnotation;
    int wBlocks;
    int hBlocks;
    
    nCoefs = pars.number[0];
    nBlocks = pars.number[1];

    if (im.w > 0 && im.h > 0) { 

        hBlocks = ceil((float)(im.h) / DCTSIZE);
        wBlocks = ceil((float)(im.w) / DCTSIZE);
        dctAnnotation = clrmalloc(wBlocks * hBlocks * nCoefs * sizeof(float));
        for (iBlk = 0; iBlk != hBlocks; iBlk++) {
            for (jBlk = 0; jBlk != wBlocks; jBlk++) {
                for (i = 0; i != DCTSIZE; i++) {
                    for (j = 0; j != DCTSIZE; j++) {
                        iPos = (i + iBlk * DCTSIZE);
                        jPos = (j + jBlk * DCTSIZE);
                        ijPosArray = ((iPos * im.w) + jPos);
                        if ((iPos > im.h) || (jPos > im.w))
    	                    block[i][j] = 0;
	                    else
        	                block[i][j] = (im.cont[ijPosArray * 3] + im.cont[ijPosArray * 3 + 1] + im.cont[ijPosArray * 3 + 2]) / 3 - 128;
                    }
                }
                blkPosArray = (iBlk * wBlocks + jBlk);
                dct_float((float *) block);
                copy_coefs((float *) block, (float *) &(dctAnnotation[blkPosArray * nCoefs]), nCoefs);
            }
        }
    } else
        return 0;

    float temp = (float) wBlocks;
    fwrite(&temp, sizeof(float), 1, fp);

    temp = (float) hBlocks;
    fwrite(&temp, sizeof(float), 1, fp);

    fwrite(dctAnnotation, sizeof(float), wBlocks * hBlocks * nCoefs, fp);
    efree(dctAnnotation);

    return 1;
}

void copy_coefs(float *zigzagBlock, float *seqBlock, int nCoefs)
{
    int i = 0, j = 0, k = 0;

    seqBlock[k++] = zigzagBlock[i * DCTSIZE + j];
    while (((i != (DCTSIZE - 1)) || (j != (DCTSIZE - 1))) && (k < nCoefs)) {
        if (((i + j) % 2) == 0) {
            j = j + 1;
            i = i - 1;
        } else if (((i + j) % 2) != 0) {
            j = j - 1;
            i = i + 1;
        }

        if (j == -1 && i != DCTSIZE) {
            j = 0;
        } else if (i == -1 && j != DCTSIZE) {
            i = 0;
        } else if (j == DCTSIZE) {
            j = 7;
            i = i + 2;
        } else if (i == DCTSIZE) {
            i = 7;
            j = j + 2;
        }
        seqBlock[k++] = zigzagBlock[i * DCTSIZE + j];
    }
}


/*******************************************************************
 * 
 * dct_float(float *data)
 * 
 * Part of the code take from jpeglib. It's an internal function, 
 * so it's not possible to call from outside the library.
 * 
 ********************************************************************/
void dct_float(float *data)
{
    double divisors[DCTSIZE] = {
        1.0, 1.387039845, 1.306562965, 1.175875602,
        1.0, 0.785694958, 0.541196100, 0.275899379
    };
    float tmp0, tmp1, tmp2, tmp3, tmp4, tmp5, tmp6, tmp7;
    float tmp10, tmp11, tmp12, tmp13;
    float z1, z2, z3, z4, z5, z11, z13;
    float *dataptr;
    int ctr, i, j;

    /* Pass 1: process rows. */

    dataptr = data;
    for (ctr = DCTSIZE - 1; ctr >= 0; ctr--) {
        tmp0 = dataptr[0] + dataptr[7];
        tmp7 = dataptr[0] - dataptr[7];
        tmp1 = dataptr[1] + dataptr[6];
        tmp6 = dataptr[1] - dataptr[6];
        tmp2 = dataptr[2] + dataptr[5];
        tmp5 = dataptr[2] - dataptr[5];
        tmp3 = dataptr[3] + dataptr[4];
        tmp4 = dataptr[3] - dataptr[4];

        /* Even part */

        tmp10 = tmp0 + tmp3;    /* phase 2 */
        tmp13 = tmp0 - tmp3;
        tmp11 = tmp1 + tmp2;
        tmp12 = tmp1 - tmp2;

        dataptr[0] = tmp10 + tmp11;     /* phase 3 */
        dataptr[4] = tmp10 - tmp11;

        z1 = (tmp12 + tmp13) * ((float) 0.707106781);   /* c4 */
        dataptr[2] = tmp13 + z1;        /* phase 5 */
        dataptr[6] = tmp13 - z1;

        /* Odd part */

        tmp10 = tmp4 + tmp5;    /* phase 2 */
        tmp11 = tmp5 + tmp6;
        tmp12 = tmp6 + tmp7;

        /* The rotator is modified from fig 4-8 to avoid extra negations. */
        z5 = (tmp10 - tmp12) * ((float) 0.382683433);   /* c6 */
        z2 = ((float) 0.541196100) * tmp10 + z5;        /* c2-c6 */
        z4 = ((float) 1.306562965) * tmp12 + z5;        /* c2+c6 */
        z3 = tmp11 * ((float) 0.707106781);     /* c4 */

        z11 = tmp7 + z3;        /* phase 5 */
        z13 = tmp7 - z3;

        dataptr[5] = z13 + z2;  /* phase 6 */
        dataptr[3] = z13 - z2;
        dataptr[1] = z11 + z4;
        dataptr[7] = z11 - z4;

        dataptr += DCTSIZE;     /* advance pointer to next row */
    }

    /* Pass 2: process columns. */

    dataptr = data;
    for (ctr = DCTSIZE - 1; ctr >= 0; ctr--) {
        tmp0 = dataptr[DCTSIZE * 0] + dataptr[DCTSIZE * 7];
        tmp7 = dataptr[DCTSIZE * 0] - dataptr[DCTSIZE * 7];
        tmp1 = dataptr[DCTSIZE * 1] + dataptr[DCTSIZE * 6];
        tmp6 = dataptr[DCTSIZE * 1] - dataptr[DCTSIZE * 6];
        tmp2 = dataptr[DCTSIZE * 2] + dataptr[DCTSIZE * 5];
        tmp5 = dataptr[DCTSIZE * 2] - dataptr[DCTSIZE * 5];
        tmp3 = dataptr[DCTSIZE * 3] + dataptr[DCTSIZE * 4];
        tmp4 = dataptr[DCTSIZE * 3] - dataptr[DCTSIZE * 4];

        /* Even part */

        tmp10 = tmp0 + tmp3;    /* phase 2 */
        tmp13 = tmp0 - tmp3;
        tmp11 = tmp1 + tmp2;
        tmp12 = tmp1 - tmp2;

        dataptr[DCTSIZE * 0] = tmp10 + tmp11;   /* phase 3 */
        dataptr[DCTSIZE * 4] = tmp10 - tmp11;

        z1 = (tmp12 + tmp13) * ((float) 0.707106781);   /* c4 */
        dataptr[DCTSIZE * 2] = tmp13 + z1;      /* phase 5 */
        dataptr[DCTSIZE * 6] = tmp13 - z1;

        /* Odd part */

        tmp10 = tmp4 + tmp5;    /* phase 2 */
        tmp11 = tmp5 + tmp6;
        tmp12 = tmp6 + tmp7;

        /* The rotator is modified from fig 4-8 to avoid extra negations. */
        z5 = (tmp10 - tmp12) * ((float) 0.382683433);   /* c6 */
        z2 = ((float) 0.541196100) * tmp10 + z5;        /* c2-c6 */
        z4 = ((float) 1.306562965) * tmp12 + z5;        /* c2+c6 */
        z3 = tmp11 * ((float) 0.707106781);     /* c4 */

        z11 = tmp7 + z3;        /* phase 5 */
        z13 = tmp7 - z3;

        dataptr[DCTSIZE * 5] = z13 + z2;        /* phase 6 */
        dataptr[DCTSIZE * 3] = z13 - z2;
        dataptr[DCTSIZE * 1] = z11 + z4;
        dataptr[DCTSIZE * 7] = z11 - z4;

        dataptr++;              /* advance pointer to next column */
    }
    for (i = 0; i < 8; i++) {
        for (j = 0; j < 8; j++) {
            data[i + j * DCTSIZE] = data[i + j * DCTSIZE] / (divisors[i] * divisors[j] * 8);
        }
    }
}


