/*********************************************************************
 Copyright (C) 2003  Imperial College London
 Multimedia Knowledge Management Group

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*********************************************************************/

/*
 * recursive.c
 *
 * Code for recursive descending; each unique file (as identified by 
 * the inode number) is only operated on once - infinite loops through
 * softlinks are spotted.
 *   
 * Stefan M Rueger
 * 
 * 12 April 2005 - Joao Magalhaes:
 * 
 *    - Added single indexing file ('-z'). Index file name is 'index.idx' 
 *      and feature file is '*.ann'.
 *    - Enable/disable recursive functionality ('-r').
 *    - DCT coefficients feature.
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <errno.h>
#include <ctype.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <dirent.h>
#include <string.h>
#include <unistd.h>

#include "annotate.h"

typedef struct _lsstruct {
    char *name;                 /* file name */
    struct stat lstat;          /* lstat(2) for file */
} LS;

static void ploughdir(LS * stats, int num);
static int tabdir(LS * lp, LS ** s_stats, char **s_names);

static int check_hash(struct stat *buf);

#define efree(p) { if((p)) free((p)), (p)=0; }

static char path[MAXPATHLEN + 1];
static char *endofpath = path;

extern int istoexcl(char *p);
extern void markunlink(char *p);

static char *ext;
extern int nexcl;
extern char **excl;

FILE *annFile = NULL;
FILE *idxFile = NULL;

char nchar='f';
char mindex[250];
char mindex2[250];
char aindex[250];


void setname(char* thename)
{
int thei=0;

	while(thename[thei] !=NULL)	
	{
	mindex[thei]=thename[thei];
	mindex2[thei]=thename[thei];
	//printf("|%c|",thename[thei]);
	thei++;
	}

}

void apply(LS * stats, int num)
{
    int l, m;

    m = ext ? strlen(ext) : 0;  // set length of file ext

    for (; num--; ++stats)
        if (S_ISREG(stats->lstat.st_mode) && (l = strlen(stats->name))
            && (m == 0 || (l >= m && !strcmp(stats->name + l - m, ext)))) {     //checks file ext
            int ti, tj, res = 0;
            FILE *fp = NULL;
            AImage im, tile, tamura;

            pars.jpgname = clrmalloc(strlen(path) + strlen(stats->name) + 1);
            sprintf(pars.jpgname, "%s%s", path, stats->name);

            for (tile.cont = im.cont = NULL, tj = 0; tj < pars.ty; tj++) {
                for (ti = 0; ti < pars.tx; ti++) {
                    if (pars.flag.s || (ti == 0 && tj == 0)) {  /* splitting or first tile */
                        pars.ftfname = ftfname(pars.jpgname, ti, tj);
                        if (annFile)
                        	fp = annFile;
                      	else if (!(fp = ftfopen(pars.ftfname)))
	                    	efree(pars.ftfname);
                    }

                    if (fp) {
                        if (!im.cont) { /* have we opened the image yet? */

                            if (pars.flag.dct)
                                im = sampleDCT(ReadImageFile(pars.jpgname), pars.number[1]);
                            else
                                im = sample(ReadImageFile(pars.jpgname), pars.destsize);

                            if (pars.useYCbCr)
                                im = RGBtoYpCbCr(im);

                            if (pars.useTamura) {
                                tamura = convertToTamura(im, pars.flag.t[0], pars.flag.t[1], pars.flag.t[2]);
                                efree(im.cont);
                                im = tamura;
                            }
                            tile.w = (im.w + pars.tx - 1) / pars.tx;
                            tile.h = (im.h + pars.ty - 1) / pars.ty;
                            if (tile.w != im.w || tile.h != im.h)
                                if (tile.w && tile.h)
                                    tile.cont = clrmalloc(tile.w * tile.h * 3 * sizeof *tile.cont);
                        }

                        if (tile.cont) {        /* need to compute a tile? */
                            unsigned char *dd, *cc;
                            int ii, jj, offj = tj * im.h / pars.ty, offi = ti * im.w / pars.tx;

                            if (offj > im.h - tile.h)
                                offj = im.h - tile.h;
                            if (offi > im.w - tile.w)
                                offi = im.w - tile.w;

                            for (dd = tile.cont, jj = offj; jj < offj + tile.h; jj++)
                                for (cc = im.cont + 3 * (im.w * jj + offi), ii = 0; ii < tile.w; ii++, cc += 3, dd += 3)
                                    dd[0] = cc[0], dd[1] = cc[1], dd[2] = cc[2];
                        }

                        res += !!pars.todo(fp, tile.cont ? tile : im);

                        if (pars.flag.s || (ti + 1 == pars.tx && tj + 1 == pars.ty)) {  /* splitting or last tile */
                            fseek(fp, 0L, SEEK_END);
                            if (ftell(fp) == 0L)        /* delete feature file if nothing was written */
                                unlink(pars.ftfname);
                            else if (!pars.flag.q)
                                printf("%s -> %s\n", pars.jpgname, pars.ftfname), fflush(stdout), fflush(stderr);

                            // JMAG: Write to the index file
                            if (pars.flag.z) {
                                fprintf(idxFile, "%s\n", pars.jpgname);
                                fflush(idxFile);
                                fflush(fp);
                            } else
	                            ftfclose(fp);
                        }
                    }

                    if (res && pars.flag.U)
                        markunlink(pars.jpgname);
                }
            }

            efree(im.cont);
            efree(tile.cont);
            efree(pars.jpgname);
            efree(pars.ftfname);
        }
}

void applyIndex()
{
    int m, len;
    int fileNameLength = 256;
    char *fileName;
	char *point_to_end;
	int thepoint=1;

    m = ext ? strlen(ext) : 0;  // set length of file ext

    fileName = (char *) malloc(fileNameLength + 1);
    while (fgets(fileName, fileNameLength, idxFile)) {

		point_to_end=strrchr(fileName,'g');
		point_to_end++;
		if(point_to_end != NULL)
		{
		*point_to_end='\0';
		}
		else
		{
		point_to_end=strrchr(fileName,'G');
		point_to_end++;
		if(point_to_end != NULL)
		{
		*point_to_end='\0';
		}

		}

        len = strlen(fileName);
		
		thepoint++;

		if(thepoint % 500 == 0)
		{
		printf("(%d) (%s) \n",thepoint,fileName);
		}

        if (fileName[len - 1] == '\n')
            fileName[len - 1] = 0;
        int ti, tj, res = 0;
        AImage im, tile, tamura;

		

        for (tile.cont = im.cont = NULL, tj = 0; tj < pars.ty; tj++) {
            for (ti = 0; ti < pars.tx; ti++) {
                if (!im.cont) { /* have we opened the image yet? */

                    if (pars.flag.dct)
                        im = sampleDCT(ReadImageFile(fileName), pars.number[1]);
                    else
                        im = sample(ReadImageFile(fileName), pars.destsize);

                    if (pars.useYCbCr)
                        im = RGBtoYpCbCr(im);

                    if (pars.useTamura) {
                        tamura = convertToTamura(im, pars.flag.t[0], pars.flag.t[1], pars.flag.t[2]);
                        efree(im.cont);
                        im = tamura;
                    }
                    tile.w = (im.w + pars.tx - 1) / pars.tx;
                    tile.h = (im.h + pars.ty - 1) / pars.ty;
                    if (tile.w != im.w || tile.h != im.h)
                        if (tile.w && tile.h)
                            tile.cont = clrmalloc(tile.w * tile.h * 3 * sizeof *tile.cont);
                }
		

                if (tile.cont) {        /* need to compute a tile? */
                    unsigned char *dd, *cc;
                    int ii, jj, offj = tj * im.h / pars.ty, offi = ti * im.w / pars.tx;

                    if (offj > im.h - tile.h)
                        offj = im.h - tile.h;
                    if (offi > im.w - tile.w)
                        offi = im.w - tile.w;

                    for (dd = tile.cont, jj = offj; jj < offj + tile.h; jj++)
                        for (cc = im.cont + 3 * (im.w * jj + offi), ii = 0; ii < tile.w; ii++, cc += 3, dd += 3)
                            dd[0] = cc[0], dd[1] = cc[1], dd[2] = cc[2];
                }
                // JMAG: Changes to write so single annotation file
                res += !!pars.todo(annFile, tile.cont ? tile : im);

                if (res && pars.flag.U)
                    markunlink(fileName);
            }
        }
        fflush(idxFile);
        efree(im.cont);
        efree(tile.cont);
    }
}


void recursive(int argc, char **argv, char *e)
{
    LS *statp;
    LS *stats;
    int fcnt, i;
    struct stat sb;
    char *names;
	int pval=0;

    ext = e;
    statp = stats = (LS *) clrmalloc(argc * sizeof(LS));
    for (fcnt = i = 0; i < argc && *argv; i++, ++argv) {
        errno = 0;
        if ((pars.flag.n ? lstat : stat) (*argv, &sb)) {
            if (warning("cannot %sstat %s", pars.flag.n ? "l" : "", *argv) == ENOENT)
                continue;
            exit(1);
        }
        if (!(nexcl && istoexcl(*argv)) && !check_hash(&sb)) {
            statp->name = *argv;
            statp->lstat = sb;
            ++statp;
            ++fcnt;
        }
    }

    // JMAG: Changes for single index file
    char *annFileName=NULL;

    if (pars.flag.z) {          // Single index file

        annFileName = ftfname("feat", pars.tx, pars.ty);
        strcat(annFileName, ".ann");

	//	pval = strlen(indexfilename);

	int i;
	
		i=0;
		while(annFileName[i] != NULL)
		{
		aindex[i]=annFileName[i];
		//printf("(%c)",aindex[i]);
		i++;
		}
		aindex[i]=NULL;
		//printf("\n\n");

		strcat(mindex,"index.idx");
		strcat(mindex2,aindex);

		//printf("|%s|%s| \n",mindex,mindex2);


		annFile = fopen(mindex2, "w");
        idxFile = fopen(mindex, "r");
        if (idxFile == NULL)
            idxFile = fopen(mindex, "w");
        else {
            fcnt = 0;
            applyIndex();
        }
    }
    // JMAG: Changes to enable/disable recursiveness
    if (fcnt) {
        if (pars.flag.r) {
            /* plough through files */
            ploughdir(stats, fcnt);
        } else if (S_ISDIR(stats->lstat.st_mode)) {
            if ((fcnt = tabdir(stats, &statp, &names))) {
                apply(statp, fcnt);
                efree(statp);
                efree(names);
            } else {
                apply(stats, fcnt);
            }
        }
    }
    fclose(annFile);
    fclose(idxFile);
    efree(annFileName);
}

/* ploughs through list of files - and then applies a recursive going through directories  */
static void ploughdir(LS * stats, int num)
{
    char *p, *savedpath;
    LS *lp, *lstats = NULL;
    int lnum;
    char *names = NULL;

    apply(stats, num);

    savedpath = endofpath;
    for (lp = stats; num--; ++lp) {
        if (!S_ISDIR(lp->lstat.st_mode))
            continue;
        p = lp->name;
        for (; (*endofpath = *p++); ++endofpath);
        if (endofpath[-1] != '/') {
            *endofpath++ = '/';
            *endofpath = 0;
        };
        if ((lnum = tabdir(lp, &lstats, &names))) {
            ploughdir(lstats, lnum);
            efree(lstats);
            efree(names);
        }
        *(endofpath = savedpath) = '\0';
    }
}

static int tabdir(LS * lp, LS ** s_stats, char **s_names)
{
    DIR *dirp;
    int cnt, maxentry, namsiz = 0;
    char *p, *names = NULL;
    struct dirent *dp;
    LS *stats = NULL;
    char fname[2 * MAXPATHLEN + 3], *namep;

    strcpy(fname, path);
    namep = strchr(fname, 0);
    strcpy(namep, ".");

    errno = 0;
    if (!(dirp = opendir(fname))) {
        warning("cannot opendir %s", fname);
        return 0;
    }

    if (pars.flag.q == 1)
        fflush(stdout), fflush(stderr), printf("%s\n", fname), fflush(stdout);
    maxentry = 0;
    for (cnt = 0; (dp = readdir(dirp));) {
        p = dp->d_name;
        if (!p)
            continue;
        if (p[0] == '.') {
            if (!p[1] || (p[1] == '.' && !p[2]))
                continue;
        }
        if (cnt == maxentry) {
            if (!maxentry)
                *s_names = names = clrmalloc(namsiz = lp->lstat.st_size + 100 * MAXPATHLEN);
            maxentry += 1024;
            *s_stats = stats = erealloc(stats, maxentry * sizeof(LS));
        }

        if (dp->d_name) {
#ifdef _DIRENT_HAVE_D_NAMLEN
            bcopy(dp->d_name, namep, (int) dp->d_namlen);
            namep[dp->d_namlen] = 0;
#else
            strcpy(namep, dp->d_name);
#endif
            if (nexcl && istoexcl(fname))
                continue;
            errno = 0;
            if ((pars.flag.n ? lstat : stat) (fname, &stats[cnt].lstat)) {
                /* don't exit - this could be an NFS mount that has gone away */
                warning("cannot %sstat %s", pars.flag.n ? "l" : "n", fname);
                continue;
            }
            if (check_hash(&stats[cnt].lstat))
                continue;
            stats[cnt].name = names;

            if (namsiz - (*s_names - names) > MAXPATHLEN) {
                strcpy(names, namep), names += strlen(names) + 1;
            } else
                *names = '\0';
        }

        ++cnt;
    }
    closedir(dirp);

    if (!cnt) {
        efree(stats);
        efree(names);
    }
    return cnt;
}

void *erealloc(void *p, u_int size)
{
    void *realloc(void *p, u_int size);

    if (!p)
        return clrmalloc(size);

    if (size == 0)
        size = 1;

    if (!(p = (void *) realloc(p, size)))
        fatal("out of memory for realloc(p,%d)", size);
    return p;
}

void *clrmalloc(u_int size)
{
    char *retval; //, *calloc();

    if (size == 0)
        size = 1;

    if (!(retval = calloc(1, size)))
        fatal("out of memory for calloc(%d)", size);
    return retval;
}

char *mystrdup(char *p)
{
    char *ret = strdup(p ? p : "");

    if (!ret)
        fatal("out of memory for strdup(\"%s\")", p ? p : "");
    return ret;
}


#define HASHSIZE 1024           /* should be power of 2 */
#define ENT_MAX 10
#define ENT_SIZE sizeof(hashent)

typedef struct {
    ino_t inode;
    off_t size;
    time_t time;
} hashent;

typedef struct hashlinestruct {
    struct hashlinestruct *next;
    int nentries;
    hashent entries[ENT_MAX];
} hashline;

static hashline hashtab[HASHSIZE];

static int check_hash(struct stat *buf)
{
    static hashent he;
    hashline *hlp;
    int i;

    if (!buf->st_ino)
        return 0;

    he.inode = buf->st_ino;
    he.size = buf->st_size;
    he.time = buf->st_ctime;

    hlp = &hashtab[buf->st_ino % HASHSIZE];
    while (hlp) {
        for (i = 0; i < hlp->nentries; i++)
            if (!memcmp(&hlp->entries[i], &he, ENT_SIZE))
                return 1;
        if (i == ENT_MAX) {
            if (hlp->next) {
                hlp = hlp->next;
                continue;
            }
            hlp = hlp->next = (hashline *) clrmalloc(sizeof(hashline));
            i = 0;
        }
        memcpy(&hlp->entries[i], &he, ENT_SIZE);
        hlp->nentries++;
        hlp = 0;
    }
    return 0;
}

//
//int getline(FILE *fp, char *buf, MAXLEN) {
//      int i;
//      
//      while ((buf[i] != '\n') && (i != MAXLEN))
//              buf[i] = getchar(fp);
//              
//      return i;
//}
//
