/*********************************************************************
 Copyright (C) 2003  Imperial College London
 Multimedia Knowledge Management Group

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*********************************************************************/

/*
 * smoothness.c
 *
 * Authors:
 * Daniel Heesch, June 2003: main program
 * Stefan Rueger, Aug 2003: slight changes to include it into annotate.c
 * 
 * ./ann -a smoothness
 * computes the smoothness of a gray image (1/(1 + scaleinvariantVariance))
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <errno.h>
#include <ctype.h>
#include <string.h>

#include "annotate.h"


#define levels (sizeof hist/sizeof*hist)

int smoothness(FILE *fp, AImage im) {
  double hist[(3*256)/8], n_hist, moms[2], x, xi;

  float smooth;
  unsigned char *c;
  int i, j;

  memset(hist, 0, sizeof hist);
  memset(moms, 0, sizeof moms);

  if(im.w>0 && im.h>0) {        /* compute gray histogram */
    for(c=im.cont, j = 0; j<im.h; j++)
      for(i = 0; i<im.w; i++, c+=3) 
        hist[(c[0]+c[1]+c[2])*levels/(3*256)]++;
    for(n_hist=im.w*im.h, i=0; i<levels; i++)
      hist[i] /= n_hist, moms[0] += hist[i]*i; /* normalise histogram & compute mean */
    for(i=0; i<levels; i++)  /* central moments */
      for(x=xi=(i-moms[0]), j=1; j<sizeof moms/sizeof*moms; j++)
        x *= xi, moms[j] += hist[i]*x;
  }

  smooth =  1 - 1/(1 + moms[1]/(levels*levels));
  fwrite(&smooth, sizeof smooth, 1, fp);

  return 1;
}
