/*********************************************************************
 Copyright (C) 2003  Imperial College London
 Multimedia Knowledge Management Group

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*********************************************************************/

/*
 * sqthumbJpg.c
 *
 * Creates a square thumbnail jpg file with extension .Jpg.
 * The aspect of the original image is kept, background is padded with gray.
 *
 * Stefan M Rueger
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <errno.h>
#include <ctype.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <string.h>

#include "annotate.h"

static int N, BGC;

int mksqthumbcheck() {
  if(pars.number[0]<1 || pars.number[0] > 2048)
    fatal("size of square thumbnail (%d) must be in range [1,2048]", pars.number[0]);
  if(pars.number[1]<0 || pars.number[1] > 255)
    fatal("gray value of background (%d) must be in range [0,255]", pars.number[1]);
  if(pars.tx >1 || pars.ty>1) /* tiled version? need to split the Jpg files */
    pars.flag.p = pars.flag.s = 1;
  pars.ftfext = ".Jpg";
  return 1;
}

int mksqthumb(FILE *fp, AImage im) {
  int i, j, ti, tj, m, nn, nx, ny, mx, my;
  unsigned char *c;
  double dx, dy;
  static float *thumb, *t;
  static int *n_thumb;
  static AImage sqtn;

  N=pars.number[0], BGC=pars.number[1];

  if(!thumb) { /* initialise */
    thumb = clrmalloc(N*N*3*sizeof*thumb);
    n_thumb = clrmalloc(N*N*sizeof*n_thumb);
    sqtn.w=sqtn.h=N;
    sqtn.cont = clrmalloc(N*N*3*sizeof*sqtn.cont);
  } else {
    memset(thumb,0,N*N*3*sizeof*thumb);
    memset(n_thumb,0,N*N*sizeof*n_thumb);
  }
  memset(sqtn.cont, BGC, N*N*3); /* grey background */

  if(im.w>0 && im.h>0) {
    dx = (double)N/im.w, dy = (double)N/im.h;
    if(dx<dy) dy=dx; else dx=dy; /* keep aspect ratio */
    if(dx>1) dy/=dx, dx=1.;      /* avoid "enlargment" */
    if(dy>1) dx/=dy, dy=1.;

    nx = dx*im.w, ny=dy*im.h;
    mx = (N-nx)/2, my=(N-ny)/2;

    for(c=im.cont, j = 0; j<im.h; j++)
      for(tj=dy*j, tj*=nx, i = 0; i<im.w; i++, c+=3) {
        ti = dx*i, ti+=tj;
        thumb[ti*3 +0] += c[0];
        thumb[ti*3 +1] += c[1];
        thumb[ti*3 +2] += c[2];
        n_thumb[ti] ++;
      }
    for(nn=nx*ny, ti=0; ti<nn; ti++)
      if((m=n_thumb[ti]))
        thumb[3*ti+0]/=m, thumb[3*ti+1]/=m, thumb[3*ti+2]/=m;

    for(t=thumb, tj=0; tj<ny; tj++)
      for(c = sqtn.cont+3*((tj+my)*N+mx), ti=0; ti<nx; ti++)
        *c++ = *t++, *c++ = *t++, *c++ = *t++;

    WriteImageFile(sqtn, fp);
  }

  return 1;
}
