/*********************************************************************
 Copyright (C) 2003  Imperial College London
 Multimedia Knowledge Management Group

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*********************************************************************/

/*
 * thumbnail.c
 *
 * Creates a thumbnail feature representation for an image.
 *
 * Stefan M Rueger
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <errno.h>
#include <ctype.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <string.h>

#include "annotate.h"


int thumbnailcheck() {
  int i;
  for(i=0; i<2; i++)
    if(pars.number[i]<=0 || pars.number[i]>2048)
      fatal("thumbnail %c-dimension (%d) is out of range [1,2048]", 'X'+i, pars.number[i]);
  
  if(strcasecmp("RGB", pars.space[0]) && strcasecmp("Y--", pars.space[0]))
    fatal("thumbnail must be ...-RGB (for colour) or ...-Y-- (for b/w) but not ...-%s", pars.space[0]);
    
  return 1;
}

int thumbnail(FILE *fp, AImage im) {
  int NX, NY, ncol;
  int i, j, ti, tj, m, nn;
  unsigned char *c;
  double dx, dy;
  static float *thumb;
  static int *n_thumb;

  NX=pars.number[0], NY=pars.number[1];
  ncol = !strcasecmp("RGB", pars.space[0])? 3: 1;

  if(!thumb) { /* initialise */
    thumb = clrmalloc(NX*NY*ncol*sizeof*thumb);
    n_thumb = clrmalloc(NX*NY*sizeof*n_thumb);
  } else {
    memset(thumb, 0, NX*NY*ncol*sizeof*thumb);
    memset(n_thumb, 0, NX*NY*sizeof*n_thumb);
  }

  if(im.w>0 && im.h>0) {
    dx = (double)NX/im.w, dy = (double)NY/im.h;
    
    if(dx<=1 && dy<=1) {        /* reduction in both directions */
      for(c=im.cont, j = 0; j<im.h; j++)
        for(tj=dy*j, tj*=NX, i = 0; i<im.w; i++, c+=3) {
          ti = dx*i, ti+=tj;
          if(ncol == 3) {
            thumb[ti*3 +0] += c[0];
            thumb[ti*3 +1] += c[1];
            thumb[ti*3 +2] += c[2];
          } else {
            thumb[ti] += (c[0]*0.299 + c[1]*0.587 + c[2]*0.114)*0.99999;
          }
          n_thumb[ti] ++;
        }
      if(ncol==3) {
        for(nn=NX*NY, ti=0; ti<nn; ti++)
          if((m=n_thumb[ti]))
            thumb[3*ti+0]/=m, thumb[3*ti+1]/=m, thumb[3*ti+2]/=m;
      } else {
        for(nn=NX*NY, ti=0; ti<nn; ti++)
          if((m=n_thumb[ti]))
            thumb[ti]/=m;
      }
    } else {                    /* enlargement in at least one direction */
      for(nn=j=0; j<NY; j++)
        for(tj=im.w*(int)(j/dy), i=0; i<NX; i++, nn+=ncol) {
          ti = (tj+(int)(i/dx))*3;
          if(ncol==3) {
            thumb[nn+0] = im.cont[ti+0];
            thumb[nn+1] = im.cont[ti+1];
            thumb[nn+2] = im.cont[ti+2];
          } else {
            thumb[nn] =  (im.cont[ti+0]*0.299 + im.cont[ti+1]*0.587 + im.cont[ti+2]*0.114)*0.99999;
          }
        }
    }
  }

  fwrite(thumb, sizeof*thumb, NX*NY*ncol, fp);

  return 1;
}
