/*********************************************************************
 Copyright (C) 2003  Imperial College London
 Multimedia Knowledge Management Group

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*********************************************************************/

/*
 * uniformity.c
 * 
 * Authors:
 * Daniel Heesch, June 2003: main program
 * Stefan Rueger, Aug 2003: slight changes to include it into annotate.c
 * 
 * ./ann -a uniformity
 * one float value: computes the uniformity of a gray image: 
 *
 *   uniformity= sum_i (h_i)**2, 
 *
 * where h(i) is a normalised 100 bins graylevel histogram.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <errno.h>
#include <ctype.h>
#include <string.h>

#include "annotate.h"


#define levels (sizeof hist/sizeof*hist)

int uniformity(FILE *fp, AImage im) {
  double hist[100], n_hist, uni=0;
  float unif;
  unsigned char *c;
  int i, j;

  memset(hist, 0, sizeof hist);

  if(im.w>0 && im.h>0) {        /* compute gray histogram */
    for(c=im.cont, j = 0; j<im.h; j++)
      for(i = 0; i<im.w; i++, c+=3) 
        hist[(c[0]+c[1]+c[2])*levels/(3*256)]++;
    for(n_hist=im.w*im.h, uni=0, i=0; i<levels; i++)
      hist[i] /= n_hist, uni += hist[i]*hist[i]; /* normalise histogram & compute uniformity */
  }

  unif =  uni;
  fwrite(&unif, sizeof unif, 1, fp);

  return 1;
}
