/*********************************************************************
 Copyright (C) 2003  Imperial College London
 Multimedia Knowledge Management Group

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*********************************************************************/

/*
 * variance.c
 * 
 * Authors:
 * Daniel Heesch, June 2003: main program
 * Stefan Rueger, Aug 2003: slight changes to include it into annotate.c
 * 
 * histogram of the standard deviation of grey-values in all TxT windows
 * ./ann -a variance-N-T   (N = number of bins, T = linear window size)
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <errno.h>
#include <ctype.h>
#include <string.h>

#include "annotate.h"


int variancecheck() {
  if(pars.number[0]<=0 || pars.number[0] >256)
    fatal("histogram size of stddeviation distribution (%d) must be in range [1,256]", pars.number[0]);
  if(pars.number[1]<=1 || pars.number[1] >10)
    fatal("window size (%d) for scanning the image must be in range [2,10]", pars.number[1]);
  return 1;
}

static int N, T;

int variance(FILE *fp, AImage im) {
  double n_hist, g, sum, ssum, stdev;
  unsigned char *g_img;
  float *s_hist;
  unsigned char *c;
  int i, j, k, bi, ci, cj, nh, nw, bin;

  N = pars.number[0];
  T = pars.number[1];

  s_hist = clrmalloc(N*sizeof*s_hist);

  if(im.w>0 && im.h>0) {        /* compute gray histogram */
    g_img = clrmalloc(im.h*im.w*sizeof*g_img);
    for(c=im.cont, k=j=0; j<im.h; j++)
      for(i=0; i<im.w; i++, c+=3)
        g_img[k++]  = (c[0]+c[1]+c[2])/3;

    for(nh=im.h-T+1, nw=im.w-T+1, j=0; j<nh; j++)
      for(i=0; i<nw; i++) {
        for(ssum=sum=0, cj=0; cj<T; cj++)
          for(bi=(j+cj)*im.w+i, ci=0; ci<T; ci++)
            g = g_img[bi+ci], sum+=g, ssum+=g*g;
        stdev = sqrt(T*T*ssum - sum*sum)/T/T;  /* standard deviation is in range [0, max/2], ie [0,127.5] */
        bin = (int)(stdev/127.5*N*.9999999);
        if(bin<0 || bin >=N)
          fatal("the standard deviation (%.f) is out of range [0, 127.5]?", stdev);
        s_hist[bin]++;
      }
    for(n_hist=nw*nh, i=0; i<N; i++)
      s_hist[i]/=n_hist;
    efree(g_img);
  }

  fwrite(s_hist, sizeof*s_hist, N, fp);
  efree(s_hist);

  return 1;
}
